#!/usr/bin/env python

"""TKDO
A to-do list aggregator with due dates, repeating events, and a snooze button.

This pulls tasks from Vim-VimOutliner (.otl) files, sorts and filters them,
and outputs a single list with tasks prioritized in order.

Based vaguely on concepts from PyGTD.
"""

# Copyright (C) 2008 Scott Scriven
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# TODO: ? rename to something better...  maybe 'masterlist' or 'vo-todo'?
program_name = "tkdo"
program_version = "0.2"

cfg = None

import re

datefmt = "%Y-%m-%d@%H:%M"

def main(args):
    """TKDO: todo list aggregator and manager for vimoutliner todo.otl files
    Usage: tkdo [global options] command [command options]
    Global options include:
      -h --help    Synonym for the "help" command.
    """
    import pycfg
    global cfg
    cfg = pycfg.config(program_name)
    set_defaults(cfg)
    create_config(cfg)
    # FIXME: tends to clobber the config data if loading had any errors
    # (which means losing all the config when it saves later)
    cfg.load()

    # TODO: keep a log of actions, for later reporting
    #   (answer 'what did I finish this week?', for example)
    # TODO: config option for save-and-exit immediately

    # handle command line
    if args:
        for i, a in enumerate(args):
            if a[0] != '-': break
            elif a in ['-h', '--help']:
                return cmd_help(*args[i+1:])
            else:
                return cmd_help()
        args = args[i:]

    if not args:
        args = [cfg.default_mode]

    op = args[0]
    try:
        func = globals()['cmd_%s' % (op)]
    except:
        print 'Invalid command: %s' % (op)
        return

    return func(*args[1:])

def set_defaults(cfg):
    cfg.Taskfile = Taskfile

    cfg.doc(active_bonus=["Score multiplier for 'active' tasks.",
        "(float, slightly higher than 1.0)"])
    cfg.default(active_bonus=1.25)

    cfg.doc(default_mode=["Which command to run, if none is specified on command line?",
        "(run 'tkdo help' to list possible commands)"])
    cfg.default(default_mode='gui')

    cfg.doc(files="A list of Taskfile objects.")
    cfg.default(files=[])

    cfg.doc(hide_zero="Should we hide zero-score items?")
    cfg.default(hide_zero=True)

    cfg.doc(hide_done="Should we hide completed tasks?")
    cfg.default(hide_done=True)

    cfg.doc(make_backups=["Copy task files to .bak before saving?",
        "If True, will copy file.otl to file.otl.bak before saving changes."])
    cfg.default(make_backups=True)

    cfg.doc(max_overdue=["Maximum score multiplier for overdue non-recurring tasks.",
        "(float, higher than 1.0)"])
    cfg.default(max_overdue=1.5)

    cfg.doc(max_overdue_fixed=["Maximum score multiplier for overdue 'fixed' recurring tasks.",
        "(float, higher than 1.0)"])
    cfg.default(max_overdue_fixed=1.5)

    cfg.doc(max_overdue_float=["Maximum score multiplier for overdue 'floating' recurring tasks.",
        "(float, higher than 1.0)"])
    cfg.default(max_overdue_float=1.25)

    cfg.doc(window_size=["Default window width and height in pixels.",
        "Should be a (x,y) tuple, such as: (640,480)"])
    cfg.default(window_size=(500,600))

    return cfg

def create_config(cfg):
    import os
    dirname = "%s/.%s" % (os.environ["HOME"], program_name)
    rcpath = "%s/rc" % (dirname)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    if not os.path.exists(rcpath):
        open(rcpath, "w")
        cfg.save()

def cmd_help(*args):
    """Display help.  (for example, "tkdo help gui")
    Usage: help [topic1 [topic2 ...]]
    If no topic is requested, an overview and list of commands is shown.
    """
    def getdoc(name):
        doc = ""
        func = globals()['cmd_%s' % (name)]
        try: doc = func.__doc__
        except: pass
        if not doc: doc = "[undocumented]"
        return doc

    if args:
        for a in args:
            doc = getdoc(a)
            print doc
    else:
        print main.__doc__
        print "Commands include:"
        fmt = "    %-10s %s"
        cmds = [k for k in globals() if k.startswith('cmd_')]
        cmds.sort()
        for c in cmds:
            name = c.replace('cmd_', '')
            text = getdoc(name).split('\n')[0].strip()
            print fmt % (name, text)

def cmd_cfg(*args):
    """Show the current configuration.
    Usage: cfg [options]
    Options include:
      -o PATH  --outpath   Write config data to PATH.
    Suggested use:
      tkdo cfg -o ~/.tkdo/rc
      vim ~/.tkdo/rc
    """
    outpath = None
    i = 0
    while i < len(args):
        a = args[i]
        if a in ['-o', '--outpath']:
            i += 1
            outpath = args[i]
        else:
            return cmd_help('cfg')
        i += 1

    if outpath:
        text = str(cfg)
        fp = open(outpath, 'wb')
        fp.write(text)
        fp.close()
    else:
        print cfg

def cmd_gui(*args):
    """Run the graphical user interface.
    Usage: gui
    (has no parameters)
    """
    #ml = Masterlist(cfg.files)
    ml = MasterlistGUI(cfg.files)
    ml.load()
    ml.sort()
    #ml.print_list()
    ml.run()

def cmd_list(*args):
    """Print the master list in plain text format.
    Usage: list [options] [search terms] [@contexts]
    Options include:
      -N              Only show the top N items.
      -r  --reverse   Print in reverse order (least-to-most important).
      -z  --zero      Show zero-score (snoozed) items.
      -Z  --no-zero   Hide zero-score (snoozed) items.
      -d  --done      Show completed items.
      -D  --no-done   Hide completed items.
    """
    reverse = False
    limit = None
    contexts = []
    filters = []
    for a in args:
        if a in ['-r', '--reverse']:
            reverse = True
        elif a in ['-z', '--zero']:
            cfg.hide_zero = False
        elif a in ['-Z', '--no-zero']:
            cfg.hide_zero = True
        elif a in ['-d', '--done']:
            cfg.hide_done = False
        elif a in ['-D', '--no-done']:
            cfg.hide_done = True
        elif a.startswith('-'):
            try:
                limit = int(a[1:])
            except:
                return cmd_help('list')
        elif a.startswith('@'):
            contexts.append(a[1:])
        else:
            filters.append(a)

    ml = Masterlist(cfg.files)
    ml.load()
    ml.sort()
    tasks = ml.filter(filters=filters, contexts=contexts)
    lines = ml.leaves_as_text(tasks)
    if limit:
        lines = lines[:limit]

    if reverse:
        lines.reverse()

    for l in lines:
        print l

def cmd_version(*args):
    """Print the program name and version.
    Usage: version
    """
    print "%s %s" % (program_name, program_version)

class Task:
    # removed 'project' concept, since I didn't use it
    valid_types = ('task', 'next', 'idea', 'ignored')
    active_types = ('task', 'next', 'idea')

    # only for recurring events, not plain old 'due' events
    recur_types = ('float', 'fixed')

    meta_pats = dict(
            type = re.compile(r"T=(\S+)"),
            # inferred from due_date
            # TODO: make HH:MM optional
            # TODO: add non-recurring due dates
            due_date = re.compile(r"D=(\d{4}-\d\d-\d\d@\d\d:\d\d)?(\+([\.\d]+)([hdwmy]))?"),
            lead_time = re.compile(r"L=([\.\d]+)([hHdDwWmMyY])"),
            snooze_time = re.compile(r"Z=(\d{4}-\d\d-\d\d@\d\d:\d\d)\+([\.\d]+)([hdwmy])"),
            last_done = re.compile(r"X=(\d)?,?(\d{4}-\d\d-\d\d@\d\d:\d\d)"),
            importance = re.compile(r"I=([\d]+)"),
            depth = re.compile(r"depth=([\d]+)"),
            )

    px_pat = re.compile(r" (\d{0,3})%($|\s)") # match completion percent
    at_pat = re.compile(r" @(\S+)") # match contexts

    def __init__(self, parent=None, lines=[], lineno=0):
        self.parent = parent
        self.lines = lines
        self.lineno = lineno

        self._reset()

        self._parse_meta()

        # create child tasks, if we're not too deep already
        depth = self.inherit('depth')
        if (self.type in self.active_types) and \
                (depth > self.level + 1): # depth is 1-based, level is 0-based
            self.load_children()

    def _reset(self):
        self.level = 0
        if self.parent:
            self.level = self.parent.level + 1

        self.tasks = []

        self.alias = None
        self.title = None
        self.active = None
        self.type = None
        self.done = None
        self.completion = None
        self.last_done = None
        self.was_done = None
        self.due_date = None
        self.recur_type = None
        self.lead_time = None
        self.snoozed = None
        self.contexts = []
        self.importance = None
        self.depth = None
        self._score = None

    def get_root(self):
        "find and return our root node"
        t = self
        while t.parent:
            t = t.parent
        return t

    def get_relation(self, rel):
        t = self
        while rel:
            if rel.startswith('t'): # this task
                rel = rel[1:]
            elif rel.startswith('p'): # parent
                rel = rel[1:]
                if t.parent:
                    t = t.parent
            elif rel.startswith('f'): # entire file
                rel = rel[1:]
                t = t.get_root()
        return t

    def load_children(self):
        indent = '\t' * (self.level + 1)
        # find next-level items
        # must be indented enough, but not too much
        # must not be our meta line
        firstlines = [num for num,line in enumerate(self.lines)
                if (line.startswith(indent))
                and (not line[len(indent):].startswith('\t'))
                and (not line[len(indent):].startswith('; TKDO: '))]

        child_tasks = []
        for tnum in range(len(firstlines)):
            start = firstlines[tnum]
            try:
                next = firstlines[tnum+1]
            except:
                next = None
            tlines = self.lines[start:next]
            t = Task(parent=self, lines=tlines, lineno=start+self.lineno)
            child_tasks.append(t)

        self.tasks = child_tasks

    def has_active_subtasks(self):
        for t in self.tasks:
            if t.type in self.active_types:
                return True
        return False

    def full_title(self):
        parts = []
        parent = self
        while parent:
            if parent.alias: text = parent.alias
            else: text = parent.title
            parts.insert(0, text)
            parent = parent.parent
        parts = [str(x) for x in parts]
        return " :: ".join(parts)

    def inherit(self, key):
        "retrieve a value from self or ancestor"
        #print "inherit(%s) %s" % (key, repr(self.full_title()))
        val = None

        parent = self
        while parent and (val is None):
            val = getattr(parent, key)
            parent = parent.parent
        #print 'inherit(): %s' % str(val)
        return val

    def inherit_list(self, key):
        "return a merged list from all ancestors"
        val = getattr(self, key)

        # collect values
        parent = self.parent
        while parent:
            pval = getattr(parent, key)
            val.extend(pval)
            parent = parent.parent

        # remove duplicates
        dv = dict(zip(val, [1]*len(val)))
        val = dv.keys()

        return val

    def _parse_meta(self):
        if len(self.lines) >= 2:
            tline, mline = self.lines[:2]
        elif len(self.lines) >= 1:
            tline = self.lines[0]
            mline = None
        else:
            tline = mline = None
        self._parse_firstline(tline)
        self._parse_metaline(mline)

    def _parse_firstline(self, line):
        if not line:
            self.type = 'ignored'
            return

        # amount of indentation; depth in outline tree
        #self.level = counttabs(line)

        # percent done
        found = self.px_pat.search(line)
        if found:
            px = found.group(1)
            if not px: px = "0"
            self.completion = float(px) / 100.0

        # remove extra stuff from title
        mangle = (
                ('[_] ', ''),
                ('[X] ', ''),
                ('[Z] ', ''), # snoozed
                (' <--', ''),
                )
        self.title = line.strip()
        for search, replace in mangle:
            self.title = self.title.replace(search, replace)

        # determine task type and state
        if ('[_] ' in line) or ('[Z] ' in line):
            self.type = 'task'
            self.done = False
        elif ('[X] ' in line):
            self.type = 'task'
            self.done = True
        else:
            self.type = 'ignored'
            self.title = line
            self.done = False

        # leave None instead of False, so inheritance will work
        if " <--" in line: self.active = True

        # also, load (and remove) contexts
        parts = self.title.split()
        self.contexts = [p[1:] for p in parts if p.startswith('@')]
        self.title = self.at_pat.sub('', self.title)

    def _parse_metaline(self, line):
        """read the line containing all the task attributes,
        assigning any discovered values to self.
        example:
            ; TKDO: T=task D=2008-04-14@17:00+1m L=1w Z=2008-04-00@12:12+1w X=1,2008-03-14@17:33 @work @online
        """
        if not line:
            return

        ## ensure line is indented once more than self
        ## (turns out it's better to let the caller ensure this is the right line)
        #if counttabs(line) != counttabs(self.lines[0]) + 1:
        #    return
        if not line.strip().startswith('; TKDO: '):
            return

        #print line

        result = {}

        # contexts
        parts = line.split()
        contexts = [x[1:] for x in parts if x.startswith('@')]
        result['contexts'] = contexts
        # merge instead of replacing
        old = dict(zip(self.contexts, [None]*len(self.contexts)))
        new = dict(zip(contexts, [None]*len(contexts)))
        new.update(old)
        self.contexts = new.keys()

        # grab what I can via regexes
        for key, pattern in self.meta_pats.iteritems():
            found = pattern.search(line)
            if found:
                value = found.groups()
                result[key] = value

        #for key, values in result.iteritems():
        #    print "%s: %s" % (key, values)

        if 'depth' in result:
            x = result['depth'][0]
            self.depth = int(x)

        if 'type' in result:
            t = result['type'][0]
            if t in self.valid_types:
                self.type = t
            if self.type == 'idea':
                self.importance = self.inherit('importance') / 2.0

        if 'importance' in result:
            x = result['importance'][0]
            x = float(x) / 100.0
            if self.type == 'idea':
                x /= 2
            self.importance = x

        if 'last_done' in result:
            x = result['last_done']
            self.was_done = self.done
            if x[0] is not None:
                self.was_done = not not int(x[0])
            self.last_done = parse_date(x[1])

            # user checked off this item since last time we looked
            #if self.done and not self.was_done:
            #    self.last_done = time.localtime()

        if 'lead_time' in result:
            r = result['lead_time']
            self.lead_time = (float(r[0]), r[1].lower())

        if 'snooze_time' in result:
            x = result['snooze_time']
            base = parse_date(x[0])
            self.snoozed = (base, float(x[1]), x[2].lower())

        if 'due_date' in result:
            r = result['due_date']
            r = (r[0], r[2], r[3])
            base = None
            self.recur_type = None
            if r[0]:
                base = parse_date(r[0])
                self.recur_type = 'due'
                if r[1]:
                    self.recur_type = 'fixed'
            if r[1]:
                self.due_date = (base, float(r[1]), r[2].lower())
                if not base:
                    self.recur_type = 'float'
            else:
                self.due_date = (base, None, None)

        """
        if self.recur_type == 'fixed':
            # TODO: magic with self.done and adjusting our due date
            # if not done:
            #   no change
            # elif done:
            #   while last completed > due date - lead time:
            #     bump due date forward one unit
            #   if last completed < due date - lead time:
            #     no change
            #   else:
            #     done = 0
        """

    def get_completion(self):
        kids = [t for t in self.tasks if t.type in self.active_types]
        if not kids:
            if self.completion is None:
                self.completion = float(self.done)
            return self.completion

        complete = sum([t.get_completion() for t in kids])
        self.completion = complete / len(kids)
        return self.completion

    def complete(self, done=True):
        import time
        self.done = done
        self.completion = 1.0
        # FIXME: notify parent
        # FIXME: complete children

        if not self.done:
            self.completion = 0.0
            return

        self.last_done = time.localtime()

        # update our due date
        if self.recur_type in self.recur_types:
            due_date = self.due_date
            last_done = time.mktime(self.last_done)
            lead = 0.0 # TODO: unsure if this is correct
            interval = due_date[1:3]
            if self.recur_type == 'float':
                # no explicit due date
                pass
            elif self.recur_type == 'fixed':
                due = due_date[0]
                while (time.mktime(due) - lead) < last_done:
                    due = date_add(due, *interval)

                foo = list(due_date)
                foo[0] = due
                self.due_date = tuple(foo)

    def snooze(self, howlong):
        """Mark this item as snoozed, starting now, for 'howlong' time.
        'howlong' should be a number and a time unit [hdwmy]
        """
        pat = re.compile(r"^\+?([\.\d]*)\s*([hHdDwWmMyY])$")

        found = pat.search(howlong)
        if not found:
            raise ValueError, "Invalid snooze length."

        num = found.group(1)
        if num:
            num = float(num)
        else:
            num = 1.0
        unit = found.group(2).lower()

        import time
        now = time.localtime()
        self.snoozed = (now, num, unit)

    def set_active(self, state=None):
        if state is None:
            # toggle
            if self.active:
                self.active = None
            else:
                self.active = True
        else: # set to True or None (not False)
            if state:
                self.active = True
            else:
                self.active = None

    def set_type(self, type_=None):
        if type_ is None:
            type_ = 'task'
        if type_ not in self.active_types:
            raise ValueError, "unknown task type"

        old = self.type
        self.type = type_

        # handle transitions
        if type_ == 'idea':
            if old != type_:
                self.importance = self.inherit('importance') / 2.0
        else:
            if (old == 'idea') and (type_ != old):
                self.importance = self.inherit('importance') * 2.0

    def set_importance(self, imp=None):
        if imp is None: self.importance = None
        else:
            imp = min(100,max(0,imp))
            self.importance = imp / 100.0
        return int(self.inherit('importance') * 100)

    def make_metaline(self):
        "generate a meta line"
        data = []

        # somewhat ignores same-as-parent attributes while saving

        # type
        if self.type != 'task':
            data.append('T=%s' % (self.type))

        # depth
        if self.depth is not None:
            data.append('depth=%s' % (self.depth))

        # importance
        # (only save if non-default, and different than parent)
        imp = self.inherit('importance')
        pimp = 0.5
        if self.parent:
            pimp = self.parent.inherit('importance')
        if self.type == 'idea': imp *= 2
        if (self.importance is not None)  \
                and  (pimp != imp) \
                and  ((imp < pimp-0.001) or (imp > pimp+0.001)):
            data.append("I=%s" % (int(imp*100)))

        # due date / recurrance info
        if self.due_date:
            x = self.due_date
            if x[0] and x[1]:
                base = fmt_date(x[0])
                data.append("D=%s+%s%s" % (base, iround(x[1]), x[2]))
            elif x[0]:
                base = fmt_date(x[0])
                data.append("D=%s" % (base))
            elif x[1]:
                data.append("D=+%s%s" % (iround(x[1]), x[2]))
            else:
                # this shouldn't happen
                self.due_date = None

        # lead time
        if self.lead_time:
            x = self.lead_time
            data.append("L=%s%s" % (iround(x[0]), x[1]))

        # last completion date
        if self.done and not self.last_done:
            data.append('X=%s,%s' % (int(self.done),fmt_date()))
        elif self.last_done:
            data.append('X=%s,%s' % (int(self.done),fmt_date(self.last_done)))

        # snooze data
        if self.snoozed:
            # TODO: omit, if snooze end time has passed
            x = self.snoozed
            base = fmt_date(x[0])
            data.append("Z=%s+%s%s" % (base, iround(x[1]), x[2]))

        # contexts
        pcontexts = []
        if self.parent: pcontexts = self.parent.inherit_list('contexts')
        done = []
        for c in self.contexts:
            if (c not in pcontexts) and (c not in done):
                data.append('@%s' % (c))
                done.append(c)

        line = ('\t' * (self.level+1)) + '; TKDO: ' + " ".join(data)
        # avoid meta lines with nothing but an X tag
        # (those just don't matter)
        if (len(data) == 1) and (data[0].startswith('X=')):
            return None

        elif data:
            return line

        return None

    def to_otl(self):
        """Return text representing self in vimoutliner .otl format.
        """
        #print "Task.to_otl(): %s" % (self.title)
        # ignore same-as-parent attributes while saving
        if self.type == 'ignored':
            text = "\n".join(self.lines)
            return text

        import time

        lines = self.lines[:]
        depth = self.level
        # build a main line
        line = self.title

        # update 'done' based on completion
        complete = self.get_completion()
        if complete > 0.99999:
            self.done = True
        else:
            self.done = False

        # choose a task marker
        active = self.active
        if self.type in self.active_types:
            # TODO: clean up this messy logic
            if self.is_snoozed():
                if self.done and (self.recur_type not in self.recur_types):
                    line = '[X] ' + line
                else:
                    line = '[Z] ' + line
            elif self.recur_type:
                if self._score is None: self.score()
                if self.done and (self.recur_type == 'due'):
                    line = '[X] ' + line
                elif self.done or (self._score < 0.001):
                    line = '[Z] ' + line
                else:
                    line = '[_] ' + line
            elif self.done:
                line = '[X] ' + line
                active = False # completed one-time tasks are never active
            else:
                line = '[_] ' + line

        # update our percent done
        found = self.px_pat.search(line)
        if found:
            x = r" %s%%\2" % (int(100*complete))
            line = self.px_pat.sub(x, line)

        # show active tasks
        if active: line = line + " <--"

        # finish up
        line = ('\t' * depth) + line
        lines[0] = line

        # add a meta line
        line = self.make_metaline()
        if line:
            lines.insert(1, line)
            lead = lines[:2]
            remaining = lines[2:]
        else:
            lead = lines[:1]
            remaining = lines[1:]

        # recurse into child tasks when depth allows
        if self.tasks:
            lines = lead
            for t in self.tasks:
                lines.append(t.to_otl())

        else:
            # FIXME: extra meta lines aren't always removed
            # (and this is a klunky approach anyway)
            # remove old/extra meta lines
            lines = lead + [l for l in remaining
                    if not l.strip().startswith('; TKDO: ')
                    or (counttabs(l) != depth+1) ]

        return "\n".join(lines)

    def is_snoozed(self):
        import time
        now = time.time()

        snoozed = self.inherit('snoozed')
        if snoozed:
            snoozed_at = snoozed[0]
            interval = snoozed[1:3]
            due = date_add(snoozed_at, *interval)
            due = time.mktime(due)
            waketime = due - (duration(*interval) / 2.0)
            if waketime > now: # not time to wake up
                return True

            if now > due: # snooze has passed entirely
                self.snoozed = None

        return False

    def timeliness(self, when=None):
        """Compute the urgency, or scheduling pressure for this task.
        Returns a float from 0.0 to a maximum urgency (normally 1.5).
        """
        # FIXME: should probably eliminate side effects
        import time
        t = 1.0
        if when:
            now = when
        else:
            now = time.time()

        # snooze hides an item for snooze_length / 2,
        # then lets it bubble up for the other half of the snooze_length.
        parent = self
        while parent: # snooze from all parents, not just the first match
            snoozed = parent.snoozed
            if snoozed:
                start = time.mktime(snoozed[0])
                # due = start + length
                due = date_add(snoozed[0], *snoozed[1:3])
                due = time.mktime(due)
                length = due - start
                window = length/2
                until = due - now
                urgency = 1.0 - (until / window)
                urgency = min(1.0, max(0.0, urgency)) # clip to range (0, 1)
                t *= urgency

            parent = parent.parent

        # due or recurring tasks
        due_date = self.inherit('due_date')
        if due_date is not None:
            interval = due_date[1:3]

            # figure out how long our ramp-up window is
            lead_time = self.inherit('lead_time')
            if lead_time:
                lead = duration(*lead_time)
            elif interval[1]:
                lead = duration(*interval) / 2.0
            else:
                lead = duration(1,'m') # default 1 month lead time

            # figure out when we were last completed
            last_done = self.inherit('last_done')
            done = self.inherit('done')
            if last_done: # has history
                last_done = time.mktime(last_done)
            elif done: # completed, but no history attached
                # assume it was completed just now
                # FIXME: write to parent here instead?
                self.last_done = time.localtime(now)
                last_done = time.mktime(self.last_done)
            else: # has never been completed
                last_done = 0.0

            # TODO: clean up this messy due date calculation code

            recur_type = self.inherit('recur_type')
            # handle plain-old-due events
            if recur_type == 'due':
                due = due_date[0]
                due = time.mktime(due)
                start = due - lead

                if start > now: # hide if it hasn't reached the lead-in time
                    t = 0.0
                else: # start < now < due :
                    until = due - now
                    urgency = 1.0 - (until / lead)
                    urgency = min(cfg.max_overdue, max(0.0, urgency)) # clip to range (0, 1.5)
                    t *= urgency

            # handle floating recurring events
            elif recur_type == 'float':
                due = last_done + duration(*interval)
                start = due - lead
                if start > now: # hide if it hasn't reached the lead-in time
                    t = 0.0
                else: # (start < now < due) or (past due)
                    until = due - now
                    urgency = 1.0 - (until / lead)
                    urgency = min(cfg.max_overdue_float, max(0.0, urgency)) # clip to range (0, 1.25)
                    t *= urgency
                    #print "urgency: %s" % (urgency)

            # handle fixed recurring events
            elif recur_type == 'fixed':
                # base = due date = original recur date
                # while due date < date last completed:
                #     due date = due date + recur interval
                # ? recur date = due date
                due = due_date[0]
                while (time.mktime(due) - lead) < last_done:
                    due = date_add(due, *interval)

                foo = list(due_date)
                foo[0] = due
                # FIXME: write to parent here instead?
                self.due_date = tuple(foo)

                due = time.mktime(due)
                start = due - lead

                if start > now: # hide if it hasn't reached the lead-in time
                    t = 0.0
                else: # start < now < due :
                    until = due - now
                    urgency = 1.0 - (until / lead)
                    urgency = min(cfg.max_overdue_fixed, max(0.0, urgency)) # clip to range (0, 1.5)
                    t *= urgency
                    #print "urgency: %s" % (urgency)

        #print "timeliness: %s" % (t)
        return t

    def score(self, when=None):
        # plain old tasks with no special attributes are omitted if completed
        if self.done and (self.type in self.active_types) \
                and (self.recur_type not in self.recur_types):
                    self._score = 0.0
        else:
            self._score = self.inherit('importance') * self.timeliness(when)
            ## quantize, to avoid the reverse order effect caused by time 
            ## passing while scoring items
            #self._score = round(self._score, 2)
        if self.inherit('active'):
            self._score *= cfg.active_bonus
        return self._score

class Taskfile(Task):
    def __init__(self, title, url, **kwargs):
        self.title = title
        self.url = url
        Task.__init__(self)
        self._reset()

    def _reset(self):
        tmp = self.title, self.url
        Task._reset(self)
        self.title, self.url = tmp

        self.parent = None # should always be None
        self.type = 'task'
        self.level = -1 # 0 is a non-indented task; we're one up from that
        self.depth = 1 # top-level tasks only
        self._changed = 0 # track whether we need to save
        if self.importance is None:
            self.importance = 0.5

    def __repr__(self):
        text = "Taskfile(%s, %s)" % (repr(self.title), repr(self.url))
        return text

    def to_otl(self):
        text = []
        # delete our meta line
        tasks = self.root_tasks
        try:
            assert(tasks[0].lines[0].startswith('; TKDO: '))
            del tasks[0]
        except:
            pass

        # make a new meta line
        # TODO: put metaline at end of file
        line = self.make_metaline()
        if line:
            text.append(line.strip())

        # add the rest 
        for t in tasks:
            text.append(t.to_otl())

        # remove trailing blank lines
        while text[-1:] == ['']:
            del text[-1]
        # ensure file ends with '\n'
        if text  and  (not text[-1].endswith('\n')):
            text.append('')

        text = "\n".join(text)
        return text

    def load(self):
        self._reset()

        fp = open_url(self.url)
        text = fp.read()
        fp.close()
        self.lines = text.split("\n")

        self.tasks = []

        # parse taskfile attributes from first line of file
        # TODO: grab meta line from either beginning or end of file
        if (len(self.lines) > 0) and self.lines[0].startswith('; TKDO: '):
            line = self.lines[0]
            self._parse_metaline(line)

        if (self.type == 'ignored') or (self.depth < 1):
            # don't try to load tasks
            return

        # find top-level items
        toplines = [num for num,line in enumerate(self.lines) if (not line.startswith('\t'))]

        root_tasks = []
        for tnum in range(len(toplines)):
            start = toplines[tnum]
            try:
                next = toplines[tnum+1]
            except:
                next = None
            tlines = self.lines[start:next]
            t = Task(parent=self, lines=tlines, lineno=start+1)
            root_tasks.append(t)
        self.root_tasks = root_tasks

        leaves = []
        def add_leaves(mytasks, parent_next=False):
            for t in mytasks:
                next = parent_next
                if t.type == 'next':
                    next = True
                if not t.has_active_subtasks():
                    leaves.append(t)
                    if parent_next and not t.done:
                        # only include the first subtask
                        return 1
                else:
                    found = add_leaves(t.tasks, next)
                    if found and parent_next:
                        # stop, if any unfinished tasks were found
                        return found
        add_leaves(root_tasks)
        self.leaves = leaves
        self.tasks = self.leaves

    def save(self):
        # TODO: move old file to .bak first
        # FIXME: recurring tasks should be marked done or not before saving
        text = self.to_otl()
        fp = open_url(self.url, "wb")
        fp.write(text)
        fp.close()

class Masterlist:
    def __init__(self, taskfiles=None):
        self.taskfiles = []
        if taskfiles:
            self.taskfiles = taskfiles[:]
        self.tasks = []
        self.filtered_tasks = []
        self.filters = []
        self.contexts = []

    def load(self):
        for tf in self.taskfiles:
            #print repr(tf)
            tf.load()
            #print "  %s lines" % (len(tf.lines))
        self.update()

    def update(self):
        self.tasks = []
        for tf in self.taskfiles:
            tasks = [t for t in tf.tasks if t.type in t.active_types]
            self.tasks.extend(tasks)
            #unfinished = [t for t in tasks if not t.done]
            #finished = [t for t in tasks if t.done]
            #active = [t for t in tasks if t.active]
            #print "  %s tasks (%s/%s active, %s done)" % (len(tasks), len(active), 
            #        len(unfinished), len(finished))

            #for t in tasks:
            ##    print "  %s %s: %s" % (int(t.score() * 100), t.parent.title, t.title)
            #    print t.to_otl()

    def sort(self):
        # freeze time while scoring...
        # otherwise, same-score scheduled items end up in reverse order
        # (because time passed between samples)
        import time
        now = time.time()
        tasks = [(t.score(now), -t.lineno, t.title, t) for t in self.tasks]
        tasks.sort()
        tasks.reverse()
        tasks = [t[-1] for t in tasks]
        self.tasks = tasks

    def filter(self, filters=[], contexts=[]):
        self.filters = filters
        self.contexts = contexts

        self.filtered_tasks = []

        # show/hide zero-score items
        # show/hide done items
        for t in self.tasks:
            if cfg.hide_done  and  t.done  and  (t._score < 0.001):
                continue
            if cfg.hide_zero  and  (not t.done)  and  (t._score < 0.001):
                continue
            #if cfg.hide_zero  and  (t._score < 0.001)  \
            #        and  (t.inherit('snoozed')  or  (t.inherit('recur_type') 
            #        in t.recur_types)):
            #    continue

            done = ""
            if t.done: done = "[X] "
            active = ""
            if t.inherit('active'): active = " <--"
            title = "%s%s%s" % (done, t.full_title(), active)

            skipped = False
            for search in self.filters: # filter by title
                if search.lower() not in title.lower():
                    skipped = True
                    break

            # filter by context
            t_contexts = " ".join([x.lower() for x in t.inherit_list('contexts')])
            for search in self.contexts:
                if search.lower() not in t_contexts:
                    skipped = True
                    break

            if not skipped:
                self.filtered_tasks.append(t)

        return self.filtered_tasks

    def leaves_as_text(self, tasks=None):
        if tasks is None:
            tasks = self.tasks

        result = []
        for t in tasks:
            active = ""
            if t.inherit('active'): active = " <--"
            done = ""
            if t.done: done = "[X] "
            result.append("%3s %s%s%s" % (int(t._score * 100), done, 
                t.full_title(), active))
        return result

class MasterlistTUI(Masterlist):
    def __init__(self, *args, **kwargs):
        Masterlist.__init__(self, *args, **kwargs)

class MasterlistGUI(Masterlist):
    def __init__(self, *args, **kwargs):
        Masterlist.__init__(self, *args, **kwargs)
        import pygtk
        pygtk.require('2.0')
        import gtk
        self.gtk = gtk

    def run(self):
        self._build_ui()
        self.update_treestore()
        self.treeview.set_cursor(0)
        try:
            self.gtk.main()
        except KeyboardInterrupt:
            pass

    def update_treestore(self, *args):
        path, column = self.treeview.get_cursor()
        self.treestore.clear()

        filtered = self.filter(filters=self.filters, contexts=self.contexts)
        for t in filtered:
            done = ""
            if t.done: done = "[X] "
            active = ""
            if t.inherit('active'): active = " <--"
            title = "%s%s%s" % (done, t.full_title(), active)

            data = [t, int(t.score()*100), title]
            branch = self.treestore.append(None, data)
            if len(t.lines) > 1:
                data = ['', '', "\n".join(t.lines[1:])]
                self.treestore.append(branch, data)

        if path:
            self.treeview.set_cursor(path)

        # set window title to indicate number of visible tasks
        num_visible = len(filtered)
        num_total = len(self.tasks)
        num_done = len([t for t in self.tasks if t.done and (t.recur_type not in t.recur_types)])

        # show current filters too
        filters = self.filters[:]
        filters.extend([('@' + x) for x in self.contexts])
        ftext = ", ".join(filters)
        self.window.set_title("%s: %s/%s tasks (%s done) in %s files [%s]" % 
                (program_name, num_visible, num_total, num_done, 
                    len(self.taskfiles), ftext))

    def _build_ui(self):
        g = self.gtk

        self.window = g.Window(g.WINDOW_TOPLEVEL)
        self.window.set_title(program_name)
        self.window.connect("destroy", self._quit)
        self.window.connect("delete_event", self._quit)
        self.window.set_size_request(220,220) # minimum size
        self.window.set_default_size(*cfg.window_size) # actual size
        self.window.connect("size-allocate", self._on_size_allocate)

        vb = g.VBox(homogeneous=False, spacing=0)
        self.window.add(vb)

        menus = self._build_menus()
        vb.pack_start(menus, False, True, 0)

        hb = g.HBox(homogeneous=False, spacing=0)
        vb.pack_start(hb)

        self.treestore = g.TreeStore(object, str, str)
        self.treeview = g.TreeView(self.treestore)
        self.treeview.set_enable_search(False)
        self.treeview.connect('key-press-event', self._tree_keypress)
        # FIXME: scrollbar can be invisible, off the right side
        # FIXME: scroll wheel doesn't work
        # FIXME: scrollbar use doesn't keep cursor onscreen
        hb.pack_start(self.treeview)
        va = self.treeview.get_vadjustment()
        scroll = g.VScrollbar(va)
        hb.pack_start(scroll, expand=False, fill=False)

        colnum = 0

        # just the widget to expand the branch
        self.expand_column = g.TreeViewColumn('')
        self.treeview.append_column(self.expand_column)
        colnum += 1

        # task score
        self.score_column = g.TreeViewColumn('score')
        self.treeview.append_column(self.score_column)
        cell = g.CellRendererText()
        self.score_column.pack_start(cell, True)
        self.score_column.add_attribute(cell, 'text', colnum)
        colnum += 1

        # task title
        self.title_column = g.TreeViewColumn('title')
        self.treeview.append_column(self.title_column)
        cell = g.CellRendererText()
        self.title_column.pack_start(cell, True)
        self.title_column.add_attribute(cell, 'text', colnum)
        colnum += 1

        # status bar area
        hb = g.HBox()
        vb.pack_start(hb, expand=False, fill=False)

        # status bar
        self._status = g.Label()
        self._status.set_justify(g.JUSTIFY_LEFT)
        self._status.set_alignment(0.0, 0.5)
        self._status.set_padding(5,5)
        hb.pack_start(self._status)

        # entry box for search, snooze, vim command emulation, etc
        self._status_entry = g.Entry()
        self._status_entry.connect('activate', self._activate_status_entry)
        hb.pack_start(self._status_entry)

        self.window.show_all()
        self._status_entry.hide()
        self.treeview.grab_focus()

    def _build_menus(self):
        "create some menu widgets"
        # FIXME: DeprecationWarning: use gtk.UIManager instead
        menu_items = (
            ('/_File', None, None, 0, '<Branch>'),
            ("/File/_Save current task's file", '<control>S', self._save_task_cb, 0, None),
            ('/File/Save _Config file', None, self._save_config_file, 0, None),
            ('/File/_Refresh', '<control>R', self._refresh, 0, None),
            ('/File/Re_load All', None, self._reload_all, 0, None),
            ('/File/_Quit', '<control>Q', self._quit, 0, None),
            ('/_Edit', None, None, 0, '<Branch>'),
            ('/Edit/_Go to (edit) taskfile', '<control>G', self._gvim_task, 0, None),
            ('/_View', None, None, 0, '<Branch>'),
            ('/View/Expand or Collapse Task', None, self._toggle_expand, 0, None),
            ('/View/_Filter list by Title', '<control>F', self._start_search, 0, None),
            ('/View/Filter list by _Context', None, self._start_context_filter, 0, None),
            ('/View/Hide\/Show snoo_zed tasks', None, self._toggle_show_snoozed,  0, None),
            ('/View/Hide\/Show _done tasks', None, self._toggle_show_done, 0, None),
            ('/_Actions', None, None, 0, '<Branch>'),
            ('/Actions/_Edit title', None, self._start_edit_title,0, None),
            ('/Actions/Snoo_ze item', '<control>Z', self._start_snooze,0, None),
            ('/Actions/Mark item as done', '<control>X', self._complete_item, 0, None),
            ('/Actions/Toggle _Active', None, self._toggle_active, 0, None),
            ('/Actions/Set _Type', '<control>T', self._start_set_type, 0, None),
            ('/Actions/Set _Importance', '<control>I', self._start_set_importance, 0, None),
            # TODO: edit contexts
            # TODO: set due date (needs a date/duration chooser)
            ('/_Help', None, None, 0, '<LastBranch>'),
            # TODO: add a dialog listing all keybindings
            #       (by key and by function)
            )
        g = self.gtk
        accel_group = g.AccelGroup()
        factory = g.ItemFactory(g.MenuBar, '<main>', accel_group)
        factory.create_items(menu_items)
        self.window.add_accel_group(accel_group)
        self.menu_factory = factory
        menubar = factory.get_widget('<main>')
        return menubar

    def _tree_keypress(self, widget, event):
        keyval = event.keyval
        name = self.gtk.gdk.keyval_name(keyval)
        string = event.string
        #print "key: %s, %s, %s" % (keyval, name, string)
        # TODO: convert keys/responses to a table lookup
        # TODO: make key bindings configurable
        try:
            if name == "g":
                self._gvim_task()
            # start a search
            elif name in ('slash', 'L'): # "L" is what I use in mutt
                self._start_search()
            elif name == 'at': # context filter
                self._start_context_filter()
            # cancel search/filter with escape
            elif name == 'Escape':
                if self.filters or self.contexts:
                    self.filters = []
                    self.contexts = []
                    self._status_entry_mode = 'search'
                    self._activate_search(None)
            elif name == 'space':
                self._toggle_expand()
            elif name == 'x':
                self._complete_item()
            elif name == 'z': # snooze task
                self._start_snooze()
            elif name == 'colon':
                self._start_vicommand()
            elif name == 'a':
                self._toggle_active()
            elif name == 't': # set task type
                self._start_set_type()
            elif name == 'i': # set task importance
                self._start_set_importance()
            elif name == 'e': # edit title
                self._start_edit_title()
            elif name == 'd': # TODO: set due date
                #self._set_due_date()
                pass
            elif name == 'k': # TODO: mark as [S] skipped
                #self._set_due_date()
                pass
        except Exception, e:
            self.set_status(str(e))
            raise

    def _get_current_task(self, *args):
        path, column = self.treeview.get_cursor()
        if not path:
            self.set_status('Select an item first.')
            return None
        data = self.treestore[path]
        t = data[0]
        if not t:
            self.set_status('Selected item is not a task.')
            return None
        return t

    def _on_size_allocate(self, widget, allocation):
        cfg.window_size = (allocation.width, allocation.height)

    def _save_task_cb(self, *args):
        t = self._get_current_task()
        if not t:
            return
        tf = t.get_root()
        self._save_task(t, "Saved taskfile: %s" % (tf.title))

    def _save_config_file(self, *args):
        self.set_status("Saving config...")
        try:
            cfg.save()
        except Exception, e:
            self.set_status(str(e))
            raise
        self.set_status("Saved config file.")

    def _save_task(self, t, msg):
        # find task file
        tf = t.get_root()
        # save
        self.set_status("Saving %s" % (tf.url))
        try:
            tf.save()
        except Exception, e:
            self.set_status(str(e))
            raise
        self.set_status("Reloading %s" % (tf.url))
        tf.load()
        self._refresh()
        self.set_status(msg)

    def gtk_flush(self, *args):
        while self.gtk.events_pending():
            self.gtk.main_iteration()

    def _refresh(self, *args):
        self.update()
        self.sort()
        self.update_treestore()

    def _reload_all(self, *args):
        for tf in self.taskfiles:
            self.set_status("Loading %s" % (tf.url))
            tf.load()
        self._refresh()
        self.set_status("Reloaded all task files.")

    def set_status(self, text):
        self._status.set_text(text)
        self.gtk_flush()

    def _get_rel(self, t, text):
        """Parse 'text' for task parent symbols.
        Returns the appropriate task, and 'text' minus the symbols.
        symbols are: t, p, f  (for task, parent, file)
        """
        rel = ''
        # FIXME: only allow one type of letter...
        # 'ptfp' should set rel to 'p' and text to 'tfp'
        while text and (text[:1] in 'tpf'):
            rel = rel + text[:1]
            text = text[1:]
        if rel:
            t = t.get_relation(rel)
        return t, text

    def _gvim_task(self, *args):
        t = self._get_current_task()
        if not t:
            return

        tf = t.get_root()
        self.edit_taskfile(tf, t)
        tf.load()
        self._refresh()

    def _complete_item(self, *args):
        t = self._get_current_task()
        if not t:
            return

        # TODO: show a brief strike-thru
        import time

        # TODO: add UI for setting item completion percent
        if t.done: # mark NOT done
            t.complete(False)
            msg = "Unchecked"
        else: # mark done
            t.complete()
            msg = "Completed"

        # save
        self._save_task(t, "%s: %s" % (msg, t.full_title()))

    def _toggle_active(self, *args):
        t = self._get_current_task()
        if not t:
            return

        # toggle active
        t.set_active()

        self._save_task(t, "Toggled Active: %s" % (t.full_title()))

    def _toggle_show_snoozed(self, *args):
        cfg.hide_zero = not cfg.hide_zero
        self._refresh()

    def _toggle_show_done(self, *args):
        cfg.hide_done = not cfg.hide_done
        self._refresh()

    def edit_taskfile(self, tf, t):
        url = tf.url
        lineno = t.lineno
        # FIXME: escape command parameters
        # FIXME: make local copy, edit, save via url abstraction
        cmd = "gvim -f +%s %s" % (lineno, str(url))
        print cmd
        import os
        self.set_status(cmd)
        os.system(cmd)
        self.set_status("Done: " + cmd)

    def _toggle_expand(self, *args):
        path, column = self.treeview.get_cursor()
        if not path:
            return
        if self.treeview.row_expanded(path):
            self.treeview.collapse_row(path)
        else:
            self.treeview.expand_row(path, open_all=False)

    def _expand_row(self, *args):
        path, column = self.treeview.get_cursor()
        if not path:
            return
        self.treeview.expand_row(path, open_all=False)

    def _collapse_row(self, *args):
        path, column = self.treeview.get_cursor()
        if not path:
            return
        self.treeview.collapse_row(path)

    def _start_entry(self, mode, label, needs_task=False, text=''):
        "open an entry box, get some text, and pass it to a handler"
        # save the current task for later, in case the handler wants it
        t = self._get_current_task()
        self._status_entry_task = t
        if needs_task and not t:
            return

        # get user input
        self.set_status(label)
        self._status_entry_mode = mode
        self._status_entry.set_text(text)
        self._status_entry.show()
        self._status_entry.grab_focus()

    def _start_search(self, *args):
        self._start_entry('search', "Search:")

    def _start_context_filter(self, *args):
        self._start_entry('context_filter', "Context:")

    def _start_snooze(self, *args):
        self._start_entry('snooze', "Snooze: (t|p+|f)?[0-9]*[hdwmy]", needs_task=True)

    def _start_vicommand(self, *args):
        self._start_entry('vicommand', "Vi emulation:")

    def _start_set_type(self, *args):
        self._start_entry('set_type', "Set type: (p+|f)? %s" % (",".join(Task.active_types)), needs_task=True)

    def _start_set_importance(self, *args):
        self._start_entry('set_importance', "Importance: (t|p+|f)?(0-100)", needs_task=True)

    def _start_edit_title(self, *args):
        title = ""
        t = self._get_current_task()
        if t: title = t.title
        self._start_entry('edit_title', "Title:", needs_task=True, text=title)

    def _activate_status_entry(self, *args):
        self._status_entry.hide()
        text = self._status_entry.get_text()
        self._status_entry.set_text('')

        # find a handler, and call it
        mode = self._status_entry_mode
        try:
            func = getattr(self, "_activate_%s" % (mode))
        except:
            raise ValueError, "invalid mode %s" % (mode)

        try:
            func(text)
        except Exception, e:
            self.set_status(str(e))
            raise

        self._status_entry_mode = None
        self.treeview.grab_focus()

    def _activate_search(self, text):
        # filter visible data
        if text:
            self.filters.append(text)
            self.set_status("Search done.")
        else:
            self.filters = []
            self.set_status("Filter removed.")
        self.update_treestore()
        self._refresh()

    def _activate_context_filter(self, text):
        # filter visible data
        if text:
            self.contexts.append(text)
            self.set_status("Context filter done.")
        else:
            self.contexts = []
            self.set_status("Filter removed.")
        self.update_treestore()
        self._refresh()

    def _activate_snooze(self, text):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        t, text = self._get_rel(t, text)

        self.set_status("Snoozing %s" % (t.full_title()))
        try:
            t.snooze(text)
        except ValueError, e:
            self.set_status(str(e))
            return

        self._save_task(t, "Snoozed %s" % (t.full_title()))

    def _activate_vicommand(self, text):
        """make a few vi commands work, like :w and :q and :e"""
        if text == "w":
            self._save_task_cb()
        elif text == "q":
            self._quit()
        #elif text == "wq": # FIXME: should save all files instead of just one?
        #    self._save_task_cb()
        #    self._quit()
        elif text == "e":
            self._reload_all()
        # allow the user to prepend ":" to some other commands
        elif text.startswith('/'):
            self._activate_search(text[1:])
        elif text.startswith('z'):
            self._activate_snooze(text[1:])
        elif text.startswith('t'):
            self._activate_set_type(text[1:])
        elif text.startswith('i'):
            self._activate_set_importance(text[1:])
        elif text.startswith('a'):
            self._toggle_active(text[1:])
        else:
            self.set_status("Unknown vi command: %s" % (text))

    def _activate_set_type(self, text):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        if not text:
            self.set_status('Set type: aborted')
            return

        # allow setting parent or file type
        if text.startswith('p') or text.startswith('f'):
            t, text = self._get_rel(t, text)

        # allow partial matches
        for choice in Task.active_types:
            if choice.startswith(text):
                text = choice
                break

        try:
            t.set_type(text)
        except ValueError, e:
            self.set_status(str(e))
            return

        self._save_task(t, "Set type (%s): %s" % (text, t.full_title()))

    def _activate_set_importance(self, text):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        t, text = self._get_rel(t, text)

        try:
            if text:
                imp = int(text)
            else:
                imp = None
            imp = t.set_importance(imp)
        except ValueError, e:
            self.set_status(str(e))
            return

        self._save_task(t, "Set I=%s on %s" % (imp, t.full_title()))

    def _activate_edit_title(self, text):
        t = self._status_entry_task
        self._status_entry_task = None
        if not t:
            self.set_status('Select a task first.')
            return

        if text:
            t.title = text
        else:
            self.set_status('Edit title: aborted.')
            return

        self._save_task(t, "Set new title: %s" % (t.full_title()))

    def _quit(self, *args):
        self.gtk.main_quit()

def counttabs(line):
    # FIXME: there has got to be a better way to do this
    i=0
    while i<len(line) and line[i] == '\t':
        i+=1
    return i

def fmt_date(when=None):
    import time
    if not when:
        when = time.localtime()
    if isinstance(when, float):
        when = time.localtime(when)
    text = time.strftime(datefmt, when)
    return text

def parse_date(text):
    import time
    date = time.strptime(text, datefmt)
    #date = time.mktime(date)

    #print "parse_date(%s): %s" % (text, repr(date))
    return date

def date_add(date, num, unit):
    import time
    units = dict(h=(1,3), d=(1,2), w=(7,2), m=(1,1), y=(1,0))
    if unit not in units:
        raise ValueError, "bad unit: %s" % (unit)
    date = list(date)
    amt, field = units[unit]
    amt *= num
    date[field] += int(round(amt))
    date = tuple(date)
    # convert to epoch and back, to get a valid tuple
    date = time.mktime(date)
    date = time.localtime(date)

    return date

def date_diff(a, b):
    import time
    at = time.mktime(a)
    bt = time.mktime(b)
    diff = at - bt
    return diff

def duration(num, unit):
    """convert a readable time length into number of seconds.
    units are: h d w m y
    Use date_add instead, if you care about issues like leap years
    or daylight savings time.  This code gets those WRONG.
    """
    units = dict(
            h=60.0 * 60,
            d=60.0 * 60 * 24,
            w=60.0 * 60 * 24 * 7,
            m=60.0 * 60 * 24 * (365.0/12),
            y=60.0 * 60 * 24 * 365,
            )
    return num * units[unit]

def iround(num):
    "convert num (float) to str, and remove trailing zeroes"
    text = str(num)
    if text[-2:] == ".0":
        text = text[:-2]
    return text

def open_url(url, mode="rb"):
    """abstract file access to make URLs and remote files work"""
    # TODO: detect/save mtime (?)
    # (could be a pain for remote files)
    def expand_home(url):
        import os
        path = url
        path = path.replace("~", os.environ["HOME"])
        path = path.replace("$HOME", os.environ["HOME"])
        return path

    if "://" not in url:
        url = "file:///" + url
    # protocol:// (?), hostname (?), :port (?), path
    pat = re.compile(r"(\w+://)?([^/:]+)?(:\d+)?/(.+)")
    found = pat.search(url)
    if not found:
        raise ValueError, "invalid URL: %s" % repr(url)

    protocol = found.group(1)[:-3]
    hostname = found.group(2)
    portnum = found.group(3)
    if portnum: portnum = portnum[1:]
    path = found.group(4)

    if protocol == "file":
        path = expand_home(path)
        if ('w' in mode) and cfg.make_backups: # make backup first
            bak = path + '.bak'
            import shutil
            shutil.copy2(path, bak)
        fp = open(path, mode)
        return fp
    # TODO: add protocol for ssh, scp, sftp, or rsync
    else:
        raise ValueError, "unsupported protocol %s" % repr(protocol)

if __name__ == "__main__":
    import sys
    main(sys.argv[1:])
